package ca.svarb.jyacl;

import java.util.ArrayList;
import java.util.List;

import ca.svarb.utils.ArgumentChecker;

/**
 * Represents an argument read from the command line.
 * An argument will be in the form of
 *  --arg=value
 * 
 * For example, an Argument constructed with the
 * string "--foo=abc" will return "foo" for the
 * name and "abc" for the value.
 */
public class Argument {

	private String name;
	private String value;
	private boolean isShortOpt=false;
	private boolean isFlag;

	public Argument(String arg) {
		ArgumentChecker.checkNulls("arg", arg);
		arg=arg.trim();
		String[] split = arg.split("=", 2);
		if ( split.length==2 ) {
			value=split[1];
		} else {
			this.isFlag=true;
		}
		setName(split[0]);
	}
	
	public Argument(String name, String value, boolean isShortOpt) {
		ArgumentChecker.checkNulls("name", name, "value", value);
		this.name=name;
		this.value=value;
		this.isShortOpt=isShortOpt;
	}

	private void setName(String name) {
		name=name.trim();
		if ( name.startsWith("--") ) {
			this.name=name.substring(2);
		} else if(name.startsWith("-")) {
			this.isShortOpt=true;
			this.name=name.substring(1);
		} else {
			throw new IllegalArgumentException("Argument must start with - or --");
		}
		if ( this.name.equals("") ) {
			throw new IllegalArgumentException("Argument has no name");
		}
	}

	public String getName() {
		return name;
	}

	public String getValue() {
		return value;
	}

	/**
	 * Convenience method to create a list of
	 * Argument objects from an array of strings.
	 * @param args
	 * @return
	 */
	public static List<Argument> createList(String[] args) {
		List<Argument> argList = new ArrayList<Argument>();
		for( String arg : args ) {
			argList.add( new Argument(arg) );
		}
		return argList;
	}

	public boolean isFlag() {
		return this.isFlag;
	}

	public boolean isShortOpt() {
		return isShortOpt;
	}

	public boolean equals(Object obj) {
		if(obj instanceof Argument) {
			Argument argument = (Argument)obj;
			if ( this.name.equals(argument.name) &&
				 this.value.equals(argument.value) && 
				 this.isShortOpt==argument.isShortOpt ) {
				return true;
			}
		}
		return false;
	}
}
