package ca.svarb.jyacl;

import java.util.Date;

import ca.svarb.utils.ArgumentChecker;

/**
 * Represents a single command line option.<p>
 * 
 * An option has a name, a return type and a prefix.
 * The option may be also be marked as mandatory.<p>
 * 
 * Supported option return types are Integer, Date and Enum and Boolean.<p>
 * 
 * The <code>getName()</code> method returns only the option
 * name without a prefix.<p>
 * 
 * The <code>asOptionString()</code> method return the option
 * name with along with the prefix.<p>
 * 
 * The <code>toString()</code> method will return the name
 * with prefix and value appropriate for the type, surrounded
 * by "[]" if the option is not mandatory.<p>
 * 
 * Examples:<p>
 * <pre>
 *     CliOption("name", String.class, true);
 *     getName():  name
 *     asOptionString():  --name
 *     toString():  --name=VALUE
 *     
 *     CliOption("id", Integer.class, false);
 *     getName():  id
 *     asOptionString():  --id
 *     toString():  [--id=INTEGER]
 *     
 *     CliOption("start", Date.class, true);
 *     getName():  start
 *     asOptionString():  --start
 *     toString():  --start=YYYY-MM-DD
 *     
 *     CliOption("help", Boolean.class, false);
 *     getName():  help
 *     asOptionString():  [--help]
 *     toString():  [--help]
 * </pre>
 *     
 */
public class CliOption {

	public static final String CLI_OPTION_PREFIX = "--";
	private static final String[] EMPTY = {""};
	private Class<?> returnType;
	private String name;
	private boolean isMandatory;
	private String[] helpText;
	private boolean isUnique;

	public CliOption(String name, Class<?> returnType, boolean isMandatory, String[] helpText) {
		ArgumentChecker.checkNulls("name",name);
		this.returnType=returnType;
		this.name=name;
		this.isMandatory=isMandatory;
		this.helpText=helpText;
		this.isUnique=false;
		if( this.helpText==null ) {
			this.helpText=EMPTY;
		}
	}

	// Create a Unique CliOption.  Unique options cannot be mandatory.
	public CliOption(String name, Class<?> returnType, String[] helpText) {
		ArgumentChecker.checkNulls("name",name);
		this.returnType=returnType;
		this.name=name;
		this.isMandatory=false;
		this.isUnique=true;
		this.helpText=helpText;
		if( this.helpText==null ) {
			this.helpText=EMPTY;
		}
	}

	public boolean isMandatory() {
		return this.isMandatory;
	}

	public boolean isUnique() {
		return this.isUnique;
	}

	public Class<?> getReturnType() {
		return this.returnType;
	}
	
	public String getName() {
		return this.name;
	}
	
	/**
	 * Option is considered equal if it has the same name and
	 * return type.  Mandatory flag is not relevant.
	 */
	public boolean equals(Object obj) {
		if(obj instanceof CliOption) {
			CliOption option = (CliOption)obj;
			if ( this.name.equals(option.name) && this.returnType==option.returnType ) {
					return true;
				}
			}
		return false;
	}
	
	public String toString() {
		StringBuffer buf=new StringBuffer();
		if(!this.isMandatory) {
			buf.append('[');
		}
		buf.append(CLI_OPTION_PREFIX);
		buf.append(getName());
		if ( !returnType.equals(Boolean.class) ) {
			buf.append('=');
			if ( returnType.equals(Integer.class)) {
				buf.append("INT");
			} else if ( returnType.equals(Date.class) ) {
				buf.append("YYYY-MM-DD");
			} else if ( returnType.isEnum() ) {
				Object[] enums = returnType.getEnumConstants();
				String delim="";
				for (Object currEnum : enums) {
					buf.append(delim);
					buf.append(currEnum.toString().toLowerCase());
					delim="|";
				}
			} else {
				buf.append("VALUE");
			}
		}
		if(!this.isMandatory) {
			buf.append(']');
		}
		return buf.toString();
	}

	public String asOptionString() {
		return CLI_OPTION_PREFIX+getName();
	}

	public String[] getHelpText() {
		return helpText;
	}
}
