package ca.svarb.utils;

import java.util.Collection;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;

/**
 * Utility to make checking method inputs easier
 * and more consistent
 *
 */
public class ArgumentChecker {

	// Package private to allow JUnit test coverage
    ArgumentChecker() {}

    /**
     * Enforce that the given input arguments are not null.
     * Call with paired name/object values,
     * e.g. if your method in class com.example.Moo is:
     * 
     *   <code>foo(String key, String value, String default)</code>
     *   
     * Of which key and value must be non-null, then the following
     * call will check for this:
     * 
     *   <code>ArgumentChecker.checkNulls("key",key,"value",value);</code>
     *  
     *  If a null value, for example, is then passed to the method,
     *  a <code>NullPointerException</code> will be thrown
     *  with the text "com.example.Moo.foo cannot be called with null value".
     *  
     * @param args
     */
    public static void checkNulls(Object... args) {
        if ( args.length%2==1 ) {
            throw new IllegalArgumentException("ArgumentChecker.checkNulls called with unpaired args");
        }
        for (int i = 0; i < args.length; i+=2) {
            if (args[i+1]==null) {
                Pair<String, String> classAndMethodNames = getClassAndMethodNames();

            	if ( classAndMethodNames.getRight()==null ) {
                   throw new NullPointerException("Null pointer passed as method argument");
                } else {
                    throw new NullPointerException(classAndMethodNames.getLeft()+"."+classAndMethodNames.getRight()+" cannot be called with null "+args[i]);
                }
            }
        }
    }

    /**
     * Enforce that given collection is not empty.  Will throw IllegalArgumentException
     * if the collection is <code>null</code>, the collection is empty, or the collection
     * contains only <code>null</code> entries.
     * @param string
     * @param objects
     */
	public static void checkForEmptyCollection(String collectionName, Collection<?> objects) {
		if(objects==null) {
            Pair<String, String> classAndMethodNames = getClassAndMethodNames();

        	if ( classAndMethodNames.getRight()==null ) {
                throw new NullPointerException("Null pointer passed as method argument");
            } else {
                throw new NullPointerException(classAndMethodNames.getLeft()+"."+classAndMethodNames.getRight()+" cannot be called with null "+collectionName);
            }
		}

		if(objects.isEmpty()) {
            Pair<String, String> classAndMethodNames = getClassAndMethodNames();

        	if ( classAndMethodNames.getRight()==null ) {
                throw new NullPointerException("Empty collection passed as method argument");
            } else {
                throw new NullPointerException(classAndMethodNames.getLeft()+"."+classAndMethodNames.getRight()+" cannot be called with empty "+collectionName);
            }
		}
		
		boolean containsOnlyNulls=true;
		for (Object object : objects) {
			if(object!=null) {
				containsOnlyNulls=false;
				break;
			}
		}
		if(containsOnlyNulls) {
            Pair<String, String> classAndMethodNames = getClassAndMethodNames();

        	if ( classAndMethodNames.getRight()==null ) {
                throw new NullPointerException("Collection argument must contain as least one non null element");
            } else {
                throw new NullPointerException("Cannot call "+classAndMethodNames.getLeft()+"."+classAndMethodNames.getRight()+" with collection "+collectionName+" containing only null elements.");
            }
		}
	}

	private static Pair<String, String> getClassAndMethodNames() {
		StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
		String className="";
		String methodName = null;

		boolean foundArgumentCheckerGetClassAndMethodNames=false;
		boolean foundArgumentCheckerCheckNulls=false;
		for (StackTraceElement stackTraceElement : stackTrace) {
			className=stackTraceElement.getClassName();
			methodName = stackTraceElement.getMethodName();
			if ( foundArgumentCheckerGetClassAndMethodNames ) {
				if ( foundArgumentCheckerCheckNulls ) {
					break;
				} else {
					if ( className!=null && className.endsWith("ArgumentChecker") ){
						foundArgumentCheckerCheckNulls=true;
					}
				}
			} else {
				if ( className!=null && className.endsWith("ArgumentChecker") ){
					foundArgumentCheckerGetClassAndMethodNames=true;
				}
			}
		}

		Pair<String, String> classAndMethodNames = new ImmutablePair<String,String>(className, methodName);
		return classAndMethodNames;
	}
}
